import numpy as np
from billions.tools.log import *


class profolio:

    def __init__(self, date, total_amount=1000000) -> None:
        self.profolio = {'freemoney': {
            'amount': total_amount,
            'price': 1,
            'buy_price': 1
        }
        }
        self.date = date

    def get_profolio(self):
        pro = list(self.profolio.keys())
        pro.remove('freemoney')
        return pro

    def buy(self, stock, amount, price):
        if price != 0:
            if self.profolio['freemoney']['amount'] < amount*price:
                log(str(self.date) + ' -- Sorry, you don\'t have enoung money')
            else:
                if stock in self.profolio.keys():
                    self.profolio[stock]['amount'] += amount
                    self.profolio[stock]['price'] = price
                    self.profolio[stock]['buy_price'] = price
                    self.profolio['freemoney']['amount'] -= amount*price
                else:
                    self.profolio[stock] = {
                        'amount': amount,
                        'price': price,
                        'buy_price': price
                    }
                    self.profolio['freemoney']['amount'] -= amount*price

    def sell(self, stock, amount, price):
        if price != 0:
            if stock not in self.profolio.keys():
                log(str(self.date) + ' -- Sorry, you don\'t have ' + stock)
            else:
                if self.profolio[stock]['amount'] < amount:
                    log(str(self.date) + ' -- Sorry, you don\'t have enough ' + stock)
                else:
                    self.profolio[stock]['amount'] -= amount
                    self.profolio[stock][price] = price
                    self.profolio['freemoney']['amount'] += amount*price
                    if self.profolio[stock]['amount'] == 0:
                        self.profolio.pop(stock)

    def profolio_report(self):
        log(str(self.date) + ' -- my profolio: ' + str(self.profolio) +
            ' -- total: ' + str(np.mean(list(self.profolio.values()))))

    def adjust_price(self, date, prices):
        self.date = date
        codes = list(self.profolio.keys())
        for code in codes:
            if code != 'freemoney':
                if prices[code] == 0:
                    self.sell(
                        code, self.profolio[code]['amount'], self.profolio[code]['price'])
                else:
                    self.profolio[code]['price'] = prices[code]

    def sell_all(self, stocks, prices):
        for stock in stocks:
            if stock in self.profolio.keys():
                self.sell(stock, self.profolio[stock]['amount'], prices[stock])
            else:
                log(str(self.date) + ' -- Sorry, you don\'t have enough ' + stock)

    def add(self, stocks, prices):
        for stock in stocks:
            self.profolio[stock] = {    
                'amount': 0,
                'price': prices[stock],
                'buy_price': prices[stock]
            }

    def add_n_sell(self, buys, sells, prices):
        self.sell_all(sells, prices)
        self.add(buys, prices)
        if self.getlen()> 0:
            self.rebalance()

    def rebalance(self):
        each = int(self.get_total() / self.getlen())
        stocks = list(self.profolio.keys())
        for stock in stocks:
            if stock == 'freemoney':
                pass
            else:
                if self.get_money(stock) > each:
                    if self.profolio[stock]['price'] != 0:
                        self.sell(stock, (self.get_money(stock) -
                                each)/self.profolio[stock]['price'], self.profolio[stock]['price'])
        for stock in stocks:
            if stock == 'freemoney':
                pass
            else:
                if self.get_money(stock) < each:
                    if self.profolio[stock]['price'] != 0:
                        self.buy(stock, (each - self.get_money(stock))/self.profolio[stock]['price'],
                                self.profolio[stock]['price'])

    def get_money(self, stock):
        money = self.profolio[stock]['amount'] * self.profolio[stock]['price']
        return money

    def get_total(self):
        moneys = 0
        for stock in self.profolio.keys():
            money = self.get_money(stock)
            moneys += money
        return moneys

    def getlen(self):
        x = -1
        for stock in self.profolio.keys():
            x += 1
        return x



from billions.tools.analysier import analysier
import pandas as pd
import numpy as np
from tqdm import tqdm

from billions.tools.profolio import *
from billions.tools.deal_time import *

# TODO: 手续费
# TODO：买入和卖出价格，当前均使用收盘价，假设能够以收盘价成交
# TODO: 支持做空（profolio中amount可以为负值）
# TODO: 加入乘子，保证交易金额为整数


class trader:
    """
    This is the parent class for all trader class. In fact, you can use it to test many strategy.

    args:
    data: data you need
    amount: money you have in your account.
    profolio: profolio type (myprofolio, euqal_profolio)
    peride: the peride you need to change your profolio
    """

    def __init__(self, data, total_amount=1000000, peride='D', price_col='close', name='strategy'):
        """
        init function
        """
        data = data.fillna(0)
        self.data = data
        self.dict_data = self.data.to_dict('index')
        self.total_amount = total_amount
        self.peride = peride
        self.price_col = price_col
        self.lable = 1
        self.name = name
        self.closes = self.data.reset_index().pivot(
            'trade_dt', 'code', self.price_col)
        self.closes = self.closes.fillna(0)
        self.closes = self.closes.to_dict('index')

        self.lables = [1]
        self.dates = np.unique(np.array(list(self.data.index))[:, 0])
        self.codes = np.unique(np.array(list(self.data.index))[:, 1])
        self.profolio = profolio(self.dates[0], total_amount)

        self.results = {}

    # mid function

    def get_trade_date(self):
        """
        **You don't want to change this**
        it can return the trade date you need.
        """
        if type(self.peride) == int:
            self.trade_dates = ndays(self.dates, self.peride)
        else:
            if self.peride == 'D':
                self.trade_dates = self.dates
            if self.peride == 'M':
                self.trade_dates = monthly(self.dates)
            if self.peride == 'Mf':
                self.trade_dates = monthly(self.dates, type='first')
            if self.peride == 'Ml':
                self.trade_dates = monthly(self.dates, type='last')

    def today_price(self, date):
        """
        it can get returns in certain date
        **You don't want to change this**
        """
        # prices = self.data.loc[date].reset_index()[['code', self.price_col]]
        # return dict(prices.set_index('code')[self.price_col])
        return self.closes[date]
        

    # set function
    def set_lable(self, lable):
        """
        change the lable
        **You don't want to change this**
        """
        if lable != -1:
            self.lable = lable

    def set_peride(self, peride):
        """
        change the peride
        **You don't want to change this**
        """
        if peride != -1:
            self.peride = peride

    def reset_profolio(self, reset):
        """
        reset the profolio
        **You don't want to change this**
        """
        if reset:
            self.profolio = profolio(
                self.dates[0], self.total_amount)

    # analysis function
    def plot(self):
        """
        return the ploter
        **You don't want to change this**
        """
        self.analysier.plot()

    def trade_summary(self):
        """
        get summary
        **You don't want to change this**
        """
        self.summaries = self.analysier.summary()
        return self.summaries

    def summary(self):
        self.plot()
        self.trade_summary()
        self.analysier.plot_summary()

    # main function
    def _run(self, lable=-1, peride=-1, reset=True):
        """
        it go over every day and get result
        **You don't want to change this**

        return:
        result: everyday profolio and amount
        """
        # set basic info
        self.set_lable(lable)
        self.set_peride(peride)
        self.reset_profolio(reset)

        # get trade date
        self.get_trade_date()

        # satrt to trade
        networths = {}
        result = {}
        for date in tqdm(self.dates, leave=False, desc="lable-" + str(self.lable)):
            prices = self.today_price(date)
            self.profolio.adjust_price(date, prices)  # adjust daily price
            if date in self.trade_dates:
                self.trade(date, prices)
            networths[date] = self.profolio.get_total()
            result[date] = self.profolio.get_profolio()
        # result = pd.DataFrame(result.values(), index=result.keys())
        # result = result.fillna({'amount':0, 'price':0, 'buy_price': 0})
        self.results[self.lable] = result
        return pd.DataFrame([networths]).T

        # result = self.results[self.lable]
        # self.networth = pd.DataFrame(result.sum(axis=1))
        # self.networth.columns = ['networth']
        # return self.networth

    def run(self, peride=-1):
        """
        run for all lable
        **You don't want to change this**
        """
        networths = []
        for lable in self.lables:
            networth = self._run(lable=lable, peride=peride, reset=True)
            networths.append(networth)

        # get networth
        self.networth = pd.concat(networths, axis=1)
        self.networth.columns = list(
            map(lambda x: 'group-' + str(x), self.lables))

        # get long short
        if len(self.lables) > 4:
            self.longshort = self.networth.copy()
            self.longshort['long/short-1'] = self.networth['group-' +
                                                           str(1)] / self.networth['group-'+str(self.number)]
            self.longshort['long/short-2'] = self.networth['group-' +
                                                           str(2)] / self.networth['group-'+str(self.number-1)]
            for i in range(1, self.number+1, 1):
                self.longshort = self.longshort.drop('group-'+str(i), axis=1)
        elif len(self.lables) > 2:
            self.longshort = self.networth.copy()
            self.longshort['long/short-1'] = self.networth['group-' +
                                                           str(1)] / self.networth['group-'+str(self.number)]
            for i in range(1, self.number+1, 1):
                self.longshort = self.longshort.drop('group-'+str(i), axis=1)
        elif len(self.lables) < 2:
            self.longshort = -1
        self.networth = (
            ((self.networth - self.networth.shift())/self.networth.shift()+1).cumprod())
        self.analysier = analysier(self)

    def trade(self, date, prices):
        pass


class group_trader(trader):
    """
    This is the parent class for all trader class. In fact, you can use it to test many strategy.

    !!! : caculator return all using close price.

    parent args:
    data: data you need
    amount: money you have in your account.
    profolio: profolio type (myprofolio, euqal_profolio)
    peride: the peride you need to change your profolio

    unique args:
    number: the number of groups
    inverse: if the factor is inversed
    """

    def __init__(self, data, total_amount=1000000, peride='D', price_col='close', name='strategy', number=5, inverse=False):
        """
        init fuction
        """
        trader.__init__(self, data, total_amount, peride, price_col, name=name)

        # self.factor = data[['factor']]

        self.inverse = not inverse
        self.number = number

        self.lables = list(range(1, self.number+1, 1))

        if self.inverse:
            self.factors = (self.data.reset_index().pivot(
                'trade_dt', 'code', 'factor')*-1).to_dict('index')
        else:
            self.factors = self.data.reset_index().pivot(
                'trade_dt', 'code', 'factor').to_dict('index')

    # main function
    def trade(self, date, prices):
        """
        trade function: change profolio to target
        """
        # get codes
        today_factor = self.factors[date]

        ind = np.where(list(pd.qcut(list(today_factor.values()),
                            self.number, labels=range(1, self.number+1, 1)) == self.lable))[0]

        codes = np.array(list(today_factor.keys()))[ind]

        # trade
        buys = []
        sells = []
        for code in self.profolio.get_profolio():
            if code not in codes:
                sells.append(code)
        for code in codes:
            if code not in self.profolio.get_profolio():
                buys.append(code)

        self.profolio.add_n_sell(buys, sells, prices)


class top_trader(trader):

    def __init__(self, data, total_amount=1000000, peride='D', price_col='close', name='strategy', top=1):
        """
        init fuction
        """
        trader.__init__(self, data, total_amount, peride, price_col, name=name)

        self.top = top
        self.factor = data[['factor']]
        self.close = data[['close']]

        self.factors = (self.data.reset_index().pivot(
            'trade_dt', 'code', 'factor')).to_dict('index')

    # main function
    def trade(self, date, prices):
        """
        trade function: change profolio to target
        """
        today_factor = self.factors[date]
        codes = list(dict(sorted(today_factor.items(),
                                 key=lambda e: e[1], reverse=True)).keys())[:self.top]

        buys = []
        sells = []
        for code in self.profolio.get_profolio():
            if code not in codes:
                sells.append(code)
        for code in codes:
            if code not in self.profolio.get_profolio():
                buys.append(code)

        self.profolio.add_n_sell(buys, sells, prices)



# 计算组合风险收益指标
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import plotly.express as px
import plotly
import pandas as pd
import numpy as np


class analysier:
    # ploter
    def __init__(self, trader):
        self.trader = trader

    def plot_networth(self):
        fig = px.line(self.trader.networth)
        fig.update_layout(title=self.get_name())
        fig.show()

    def plot_long_short(self):
        fig = px.line(self.trader.longshort)
        fig.update_layout(title=self.get_name())
        fig.show()

    def plot(self):
        if type(self.trader.longshort) == int:
            self.plot_networth()
        else:
            fig = plotly.subplots.make_subplots(rows=1, cols=2)

            for column in self.trader.networth.columns:
                fig.append_trace(go.Scatter(
                    x=self.trader.networth.index, y=self.trader.networth[column], name=column), 1, 1)
            for column in self.trader.longshort.columns:
                fig.append_trace(go.Scatter(
                    x=self.trader.longshort.index, y=self.trader.longshort[column], name=column), 1, 2)
            fig.update_layout(title=self.get_name() + ": networth and long/short plot")
            fig.show()

    def get_name(self):
        if type(self.trader.peride) == str:
            if self.trader.peride == 'D':
                return self.trader.name + ' -- daily backtest'
            if self.trader.peride[0] == 'M':
                return self.trader.name + ' -- monthly backtest'
        if type(self.trader.peride) == int:
            return self.trader.name + ' -- ' + str(self.trader.peride) + ' days backtest'

    # summary
    def getReturn(self, nw):
        return (nw - nw.shift()) / nw.shift()

    def summary(self):
        nw = self.trader.networth
        ret = self.getReturn(nw)
        summaries = []
        for column in ret.columns:
            summary = self.tradeSummary(list(ret.index), list(
                ret[column]), column, freq='D')
            summaries.append(summary)
        if type(self.trader.longshort) != int:
            lsnw = self.trader.longshort
            lsret = self.getReturn(lsnw)
            for column in lsret.columns:
                summary = self.tradeSummary(list(lsret.index), list(
                    lsret[column]), column, freq='D')
                summaries.append(summary)
        self.summaries = pd.concat(summaries, axis=1)
        return self.summaries

    def plot_summary(self):
        df = self.summaries
        values = [df.index]
        for column in df.columns:
            values.append(df[column])
        fig = go.Figure(data=[go.Table(
            header=dict(values=[""] + list(df.columns),
                        fill_color='paleturquoise',
                        align='left'),
            cells=dict(values=values,
                       fill_color='lavender',
                       align='left'))
        ])
        fig.update_layout(title=self.get_name() + ": trade summary table")
        fig.show()

        # return summaries

    def tradeSummary(self, date, ret, group, freq='D'):
        df = pd.DataFrame({'date': date, 'ret': ret})
        df.sort_values(by='date', inplace=True)
        df.reset_index(drop=True, inplace=True)
        df.dropna(inplace=True)

        # 计算复利
        df['compound'] = np.cumprod(df['ret'] + 1)

        # 计算最大回撤
        df['max2here'] = df['compound'].expanding(1).max()
        df['dd2here'] = (df['compound'] / df['max2here']) - 1
        # df.to_excel("test.xlsx")

        # 计算最大回测开始和结束时间
        temp = df.sort_values(by='dd2here').iloc[0][['date', 'dd2here']]
        max_dd = temp['dd2here']
        end_dd = temp['date']
        start_dd = df[df['date'] <= end_dd].sort_values(
            by='compound', ascending=False).iloc[0]['date']

        # 计算判断数据频率
        if freq == 'D':
            n = 250
        elif freq == 'W':
            n = 52
        elif freq == 'M':
            n = 12
        else:
            print('请输入正确的时间频率，月：‘M’ 周：‘W’ 日：‘D’')

        rng = len(date)

        # 年化收益率
        annual_ret = pow(df['compound'].iloc[-1], n / rng) - 1

        # 波动率
        volatility = df['ret'].std() * np.sqrt(n)
        sharpe = annual_ret / volatility

        # 计算calmer
        calmar = annual_ret / abs(max_dd)

        return_risk_df = pd.DataFrame()
        return_risk_df["年化收益率"] = [str(round(annual_ret * 100, 2)) + '%']
        return_risk_df["年化波动率"] = [str(round(volatility * 100, 2)) + '%']
        return_risk_df["夏普比率"] = [round(sharpe, 3)]
        return_risk_df["最大回撤"] = [str(round(max_dd * 100, 2)) + '%']
        return_risk_df["calmar"] = [round(calmar, 3)]
        return_risk_df["回撤起始"] = [str(start_dd)[:10]]
        return_risk_df["回撤结束"] = [str(end_dd)[:10]]
        return_risk_df = return_risk_df.T
        return_risk_df.columns = [group]
        return return_risk_df
